// Listing9-6.S
//
// 32-bit ARM code for the
// decimal string to numeric conversion.
//
// Written to run on a Raspberry Pi.

#include    "ao32aa.inc"

// Number of iterations used by timing code:

#define iterations 100000000

// The following are required to use ARM V7&8 instructions
// on Raspberry Pi devices (3, 4, & 5):

            .arch   armv8-a

// Manifest constant for tab character:

#define tab 0x9

        .data
    
// Data to convert:
    
in0:    .asciz  "0"
in1:    .asciz  "1"
in2:    .asciz  "22"
in3:    .asciz  "333"
in4:    .asciz  "4444"
in4a:   .asciz  "1234"
in4b:   .asciz  "5678"
in5:    .asciz  "55555"
in5a:   .asciz  "12345"
in5b:   .asciz  "67890"
in6:    .asciz  "666666"
in7:    .asciz  "7777777"
in8:    .asciz  "88888888"
in9:    .asciz  "999_999_999"
in10:   .asciz  "4_294_967_295"
in11:   .asciz  " ;:,\t 12345678 "
in12:   .asciz  "1_999_999_999"

// Pointers to strings, so asmMain can process
// all of them in a loop:
        
strPtrs:
        .word   in0, in1, in2, in3
        .word   in4, in4a, in4b
        .word   in5, in5a, in5b
        .word   in6, in7
        .word   in8, in9, in10, in11, in12
        .set    inCnt, (.-strPtrs)/4    //Number of strings

        .text
        .align  2

        #include    "print.inc"        

// getTitle returns the address of a string in
// the R0 register. The c.cpp program will print
// this string before and after calling asmMain.

        .text
        .align  2
        
// Here's the title string that c.cpp will print:

title:  wastr  "Listing9-6"
        
        proc    getTitle, public
        adr     r0, title
        bx      lr
        endp    getTitle

////////////////////////////////////////////////////
//
// bufToU32
//
// Decimal string to 32-bit integer conversion.
// This code allows underscores to appear within a
// decimal string (but underscores cannot be the first
// character of a hex string). This code also
// allows an underscore at the end of the string;
// if you consider this to be an error, modify
// the code at label "delimiter" to test for an
// underscore in the previous character position.
//
// On entry:
//
//  R0-     Pointer to buffer holding string.
//
// On exit-
//
//  R0-     Points at the first character beyond
//          the converted string, if conversion
//          was successful.
//
//  R1-     Contains the converted value if the
//          conversion was successful.
//
//  Carry-  Clear if conversion was successful.
//          Set if unsuccessful. If unsuccessful
//          conversion, then R0 will point at
//          the character the conversion failed
//          on and R1 will contain one of the
//          following error codes:
//
//          -1: First (non-WS) character was not
//              a valid hexadecimal digit.
//
//          -2: Overflow (value did not fit into
//              32 bits).
//
//          -3: Last character was not a valid
//              delimiter character (this could
//              be considered a "soft" error).
//
//  This function will skip any leading whitespace
// (spaces or tabs) appearing at the beginning
// of the string.
        
    proc    bufToU32, thumb, public
    push    {r2-r4, lr}
    
    movs    r1, #0          // Initialize result.
    ldr     r3, =429496729  // For overflow check
    movs    r4, #10         // Used for radix multiply.
    
    // Skip any leading whitespace, comma, semicolon,
    // or colon characters in the string:
    
    subs    r0, r0, #1      // To counteract pre-inc, below
skipDelim:
    ldrb    r2, [r0, #1]!
    
    cmp     r2, #' '
    beq     skipDelim
    cmp     r2, #tab
    beq     skipDelim
    cmp     r2, #','
    beq     skipDelim
    cmp     r2, #':'
    beq     skipDelim
    cmp     r2, #';'
    beq     skipDelim
    
    // Verify that the first character is a valid
    // hexadecimal digit:
    
    cmp     r2, #'0'
    blo     badFirst
    cmp     r2, #'9'
    bls     decDigit
        
    // If the first digit is bad, return -1:
    
badFirst:
    mov     r1, #-1
    cmp     r1, r1              // Sets carry flag
    pop     {r2-r4, pc} 


overflow:
    mov     r1, #-2
    cmp     r1, r1              // Sets carry flag
    pop     {r2-r4, pc}
    
    // Process decimal digits:
    
decDigit:
    and     r2, r2, #0xf        // Convert to 0x0..0x9
    cmp     r1, r3              // Check for overflow
    bhi     overflow
    muls    r1, r1, r4          // Accumulator * 10
    adds    r1, r1, r2          // Merge into result
    bcs     overflow            // Also check for overflow

nextChar:
    ldrb    r2, [r0, #1]!
    cmp     r2, #'0'
    blo     tryDelimiter
    cmp     r2, #'9'
    bls     decDigit
        
tryDelimiter:
    cmp     r2, #'_'
    beq     nextChar

    cmp     r2, #0
    beq     delimiter
    cmp     r2, #' '
    beq     delimiter
    cmp     r2, #tab
    beq     delimiter
    cmp     r2, #','
    beq     delimiter
    cmp     r2, #':'
    beq     delimiter
    cmp     r2, #';'
    beq     delimiter
    
    // If the first non-hex-digit character
    // is bad, return -3:
    
illegal:
    mov     r1, #-3
    cmp     r1, r1              // Sets carry flag
    pop     {r2-r4, pc}

    // Transfer here when encountering a legitimate
    // delimiter character (valid end of hex string).
    
delimiter:
    adds    r0, r0, #0          // Clears carry flag
    pop     {r2-r4, pc}
    endp    bufToU32

/////////////////////////////////////////////////////////////////////
//
// strToU32
//
// Converts the string pointed at by R0 to a numeric value
// and returns that numeric value in R1. This function
// preserves R0 (leaves it pointing at the beginning of the
// string buffer).

    proc    strToU32, thumb, public
    push    {r0, lr}
    bl      bufToU32
    pop     {r0, pc}
    endp    strToU32
            
/////////////////////////////////////////////
//              
// Here's the assembly language main program:
        
    proc    asmMain, public
    push    {r0-r8, lr}

// Timing code:

    ldr     r8, =iterations
timingLp:

    ldr     r7, =strPtrs
    mov     r6, #inCnt
innerLp:     
    ldr     r0, [r7], #4
       
    // Convert string pointed at by R0 to 
    // a numeric value in R1:
    
    bl      bufToU32
            
    // Repeat for each sample input:
      
    subs    r6, #1
    bne     innerLp            

    // Repeat a large number of times in order
    // to time the code:
    
    subs    r8, r8, #1
    bne     timingLp

// Display output for testing purposes:

    ldr     r7, =strPtrs
    mov     r6, #inCnt
convLp:     
    ldr     r0, [r7], #4
       
    // Convert string pointed at by R0 to 
    // a numeric value in R1 and output
    // that value:
    
    bl      strToU32
    
    bl      print
    wastr   "Converted '%s' to %u\n"
          
    subs    r6, #1
    bne     convLp            
             
    pop     {r0-r8, pc} 
    endp    asmMain


